from lark import Lark, InlineTransformer
from mel_ast import *

parser = Lark('''
    %import common.NUMBER
    %import common.CNAME
    %import common.NEWLINE
    %import common.WS

    %ignore WS

    COMMENT: "/*" /(.|\\n|\\r)+/ "*/"
        |  "//" /(.)+/ NEWLINE
    %ignore COMMENT

    num: NUMBER
    ident: CNAME

    ?group: num
        | ident
        | "(" add ")"

    ?mult: group
        | mult "*" group    -> mul
        | mult "/" group    -> div

    ?add: mult
        | add "+" mult      -> add
        | add "-" mult      -> sub

    ?expr: add

    if:  "if" "(" expr ")" stmt ("else" stmt)?

    if2:  "if" "(" expr ")" stmt "else" stmt  -> if
        | "if" "(" expr ")" stmt              -> if

    while: "while" "(" expr ")" stmt

    ?expr_or_empty: expr
        | -> true
    ?stmt_or_empty: stmt
        | -> stmt_list
    for: "for" "(" stmt_or_empty ";" expr_or_empty ";" stmt_or_empty ")" stmt

    ?stmt: "input" ident     -> input
        | "output" expr     -> output
        | ident "=" expr    -> assign
        | if2
        | while
        | for
        | "{" stmt_list "}"

    stmt_list: stmt*

    ?prog: stmt_list

    ?start: prog
''', start="start")


class MelASTBuilder(InlineTransformer):
    def __getattr__(self, item):
        if isinstance(item, str) and item.upper() == item:
            return lambda x: x

        if item == 'true':
            return lambda: NumNode(1)
        if item in ('mul', 'div', 'add', 'sub'):
            def get_bin_op_node(*args):
                op = BinOp[item.upper()]
                return BinOpNode(op, *args)

            return get_bin_op_node
        else:
            def get_node(*args):
                cls = eval(''.join(x.capitalize() or '_' for x in item.split('_')) + 'Node')
                return cls(*args)

            return get_node


def parse(prog: str) -> StmtListNode:
    prog = parser.parse(str(prog))
    prog = MelASTBuilder().transform(prog)
    return prog
